from __future__ import annotations as _annotations

import asyncio
from types import SimpleNamespace
from typing import Any, cast

import pytest
from pydantic_ai.messages import ModelResponse, ToolCallPart
from typing_extensions import Sentinel

from .support import (
    UTC,
    AcpSessionContext,
    Agent,
    ExternalHookEventBridge,
    HookBridge,
    HookEvent,
    HookProjectionMap,
    Path,
    TestModel,
    ToolCallProgress,
    ToolCallStart,
    ToolDefinition,
    datetime,
)

_HOOK_CONTEXT = Sentinel("_HOOK_CONTEXT")


def test_hook_bridge_error_paths_emit_failed_updates() -> None:
    bridge = HookBridge()
    session = AcpSessionContext(
        session_id="hook-errors",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    capability = bridge.build_capability(session)
    registry = capability._registry
    tool_call = ToolCallPart("echo", {"text": "hello"})
    tool_def = ToolDefinition(name="echo")

    async def fail_run() -> Any:
        raise RuntimeError("run failed")

    async def fail_node(node: Any) -> Any:
        del node
        raise RuntimeError("node failed")

    async def fail_model(request_context: Any) -> ModelResponse:
        del request_context
        raise RuntimeError("model failed")

    async def fail_validate(args: Any) -> Any:
        del args
        raise RuntimeError("validate failed")

    async def fail_execute(args: Any) -> Any:
        del args
        raise RuntimeError("execute failed")

    with pytest.raises(RuntimeError, match="run failed"):
        asyncio.run(registry["wrap_run"][0].func(cast("Any", None), handler=fail_run))
    with pytest.raises(RuntimeError, match="run direct"):
        asyncio.run(
            registry["on_run_error"][0].func(cast("Any", None), error=RuntimeError("run direct")),
        )
    with pytest.raises(RuntimeError, match="node failed"):
        asyncio.run(
            registry["wrap_node_run"][0].func(
                cast("Any", None),
                node=_HOOK_CONTEXT,
                handler=fail_node,
            ),
        )
    with pytest.raises(RuntimeError, match="model failed"):
        asyncio.run(
            registry["wrap_model_request"][0].func(
                cast("Any", None),
                request_context=_HOOK_CONTEXT,
                handler=fail_model,
            ),
        )
    with pytest.raises(RuntimeError, match="validate failed"):
        asyncio.run(
            registry["wrap_tool_validate"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                handler=fail_validate,
            ),
        )
    with pytest.raises(RuntimeError, match="execute failed"):
        asyncio.run(
            registry["wrap_tool_execute"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                handler=fail_execute,
            ),
        )
    with pytest.raises(RuntimeError, match="execute direct"):
        asyncio.run(
            registry["on_tool_execute_error"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                error=RuntimeError("execute direct"),
            ),
        )

    updates = bridge.drain_updates(session, Agent(TestModel(custom_output_text="unused")))
    assert updates is not None
    failed_titles = [
        update.title
        for update in updates
        if isinstance(update, ToolCallProgress) and update.status == "failed"
    ]
    assert failed_titles == [
        "hook.wrap_run",
        "hook.on_run_error",
        "hook.wrap_node_run",
        "hook.wrap_model_request",
        "hook.wrap_tool_validate",
        "hook.wrap_tool_execute",
        "hook.on_tool_execute_error",
    ]


def test_hook_bridge_success_paths_and_disabled_metadata() -> None:
    session = AcpSessionContext(
        session_id="hook-success",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    bridge = HookBridge()
    capability = bridge.build_capability(session)
    registry = capability._registry
    tool_call = ToolCallPart("echo", {"text": "hello"})
    tool_def = ToolDefinition(name="echo")
    response = ModelResponse(parts=[])

    async def ok_run() -> Any:
        return cast("Any", "run-result")

    async def ok_node(node: Any) -> Any:
        return node

    async def ok_model(request_context: Any) -> ModelResponse:
        del request_context
        return response

    async def ok_validate(args: Any) -> Any:
        return args

    async def ok_execute(args: Any) -> Any:
        return {"ok": args["text"]}

    async def ok_output(output: Any) -> Any:
        return output

    async def fail_output(output: Any) -> Any:
        del output
        raise RuntimeError("output failed")

    async def event_stream():
        yield cast("Any", SimpleNamespace(event_kind="demo_event"))

    request_context = cast("Any", SimpleNamespace(messages=[1, 2]))
    assert asyncio.run(registry["before_run"][0].func(cast("Any", None))) is None
    assert (
        asyncio.run(registry["wrap_run"][0].func(cast("Any", None), handler=ok_run)) == "run-result"
    )
    assert (
        asyncio.run(registry["after_run"][0].func(cast("Any", None), result=cast("Any", "done")))
        == "done"
    )
    assert (
        asyncio.run(registry["before_node_run"][0].func(cast("Any", None), node=tool_call))
        is tool_call
    )
    assert (
        asyncio.run(
            registry["after_node_run"][0].func(
                cast("Any", None),
                node=tool_call,
                result=cast("Any", "node-result"),
            ),
        )
        == "node-result"
    )
    assert (
        asyncio.run(
            registry["wrap_node_run"][0].func(
                cast("Any", None),
                node=tool_call,
                handler=ok_node,
            ),
        )
        is tool_call
    )
    assert (
        asyncio.run(registry["_on_event"][0].func(cast("Any", None), cast("Any", tool_call)))
        is tool_call
    )
    wrapped_stream = registry["wrap_run_event_stream"][0].func(
        cast("Any", None),
        stream=event_stream(),
    )
    assert asyncio.run(wrapped_stream.__anext__()).event_kind == "demo_event"
    assert (
        asyncio.run(registry["before_model_request"][0].func(cast("Any", None), request_context))
        is request_context
    )
    assert (
        asyncio.run(
            registry["wrap_model_request"][0].func(
                cast("Any", None),
                request_context=request_context,
                handler=ok_model,
            ),
        )
        is response
    )
    assert (
        asyncio.run(
            registry["after_model_request"][0].func(
                cast("Any", None),
                request_context=request_context,
                response=response,
            ),
        )
        is response
    )
    assert asyncio.run(registry["prepare_tools"][0].func(cast("Any", None), [tool_def])) == [
        tool_def,
    ]
    assert asyncio.run(
        registry["before_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["after_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["wrap_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
            handler=ok_validate,
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["before_tool_execute"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["wrap_tool_execute"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
            handler=ok_execute,
        ),
    ) == {"ok": "hello"}
    assert (
        asyncio.run(
            registry["after_tool_execute"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                result="done",
            ),
        )
        == "done"
    )

    updates = bridge.drain_updates(session, Agent(TestModel(custom_output_text="unused")))
    assert updates is not None
    assert len([u for u in updates if isinstance(u, ToolCallProgress)]) >= 15

    disabled = HookBridge(
        record_event_stream=False,
        record_model_requests=False,
        record_node_lifecycle=False,
        record_deferred_tool_calls=False,
        record_output_processing=False,
        record_output_validation=False,
        record_prepare_output_tools=False,
        record_prepare_tools=False,
        record_run_lifecycle=False,
        record_tool_execution=False,
        record_tool_validation=False,
    )
    disabled_capability = disabled.build_capability(session)
    assert disabled_capability._registry == {}
    assert disabled.get_session_metadata(session, Agent(TestModel())) == {"events": []}

    partially_disabled = HookBridge()
    partially_capability = partially_disabled.build_capability(session)
    partially_registry = partially_capability._registry
    partially_disabled.record_run_lifecycle = False
    partially_disabled.record_node_lifecycle = False
    partially_disabled.record_event_stream = False
    partially_disabled.record_model_requests = False
    partially_disabled.record_deferred_tool_calls = False
    partially_disabled.record_output_processing = False
    partially_disabled.record_output_validation = False
    partially_disabled.record_prepare_output_tools = False
    partially_disabled.record_prepare_tools = False
    partially_disabled.record_tool_validation = False
    partially_disabled.record_tool_execution = False

    assert asyncio.run(partially_registry["before_run"][0].func(cast("Any", None))) is None
    assert (
        asyncio.run(
            partially_registry["before_node_run"][0].func(cast("Any", None), node=tool_call),
        )
        is tool_call
    )
    assert (
        asyncio.run(
            partially_registry["_on_event"][0].func(
                cast("Any", None),
                cast("Any", SimpleNamespace(event_kind="silent")),
            ),
        ).event_kind
        == "silent"
    )
    assert (
        asyncio.run(
            partially_registry["before_model_request"][0].func(cast("Any", None), request_context),
        )
        is request_context
    )
    assert asyncio.run(
        partially_registry["prepare_tools"][0].func(cast("Any", None), [tool_def]),
    ) == [tool_def]
    assert asyncio.run(
        partially_registry["prepare_output_tools"][0].func(cast("Any", None), [tool_def]),
    ) == [tool_def]
    assert (
        asyncio.run(
            partially_registry["before_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
            ),
        )
        == "raw"
    )
    assert (
        asyncio.run(
            partially_registry["wrap_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=ok_output,
            ),
        )
        == "raw"
    )
    with pytest.raises(RuntimeError, match="output failed"):
        asyncio.run(
            partially_registry["wrap_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=fail_output,
            ),
        )
    assert (
        asyncio.run(
            partially_registry["after_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="checked",
            ),
        )
        == "checked"
    )
    with pytest.raises(RuntimeError, match="validate failed"):
        asyncio.run(
            partially_registry["on_output_validate_error"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                error=RuntimeError("validate failed"),
            ),
        )
    assert (
        asyncio.run(
            partially_registry["before_output_process"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
            ),
        )
        == "raw"
    )
    assert (
        asyncio.run(
            partially_registry["wrap_output_process"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=ok_output,
            ),
        )
        == "raw"
    )
    with pytest.raises(RuntimeError, match="output failed"):
        asyncio.run(
            partially_registry["wrap_output_process"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=fail_output,
            ),
        )
    assert (
        asyncio.run(
            partially_registry["after_output_process"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="processed",
            ),
        )
        == "processed"
    )
    with pytest.raises(RuntimeError, match="process failed"):
        asyncio.run(
            partially_registry["on_output_process_error"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                error=RuntimeError("process failed"),
            ),
        )
    assert (
        asyncio.run(
            partially_registry["handle_deferred_tool_calls"][0].func(
                cast("Any", None),
                requests=cast("Any", None),
            ),
        )
        is None
    )
    assert asyncio.run(
        partially_registry["before_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        partially_registry["before_tool_execute"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert partially_disabled.drain_updates(session, Agent(TestModel())) is None


def test_hook_bridge_hide_all_disables_registry_and_metadata() -> None:
    bridge = HookBridge(hide_all=True)
    session = AcpSessionContext(
        session_id="hook-hidden",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )

    capability = bridge.build_capability(session)

    assert capability._registry == {}
    assert bridge.get_session_metadata(session, Agent(TestModel())) == {"events": []}
    assert bridge.drain_updates(session, Agent(TestModel())) is None


def test_hook_bridge_records_output_and_deferred_hooks() -> None:
    bridge = HookBridge()
    session = AcpSessionContext(
        session_id="hook-output",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    registry = bridge.build_capability(session)._registry
    tool_def = ToolDefinition(name="result")

    async def ok_output(output: Any) -> Any:
        return output

    assert asyncio.run(registry["prepare_output_tools"][0].func(cast("Any", None), [tool_def])) == [
        tool_def,
    ]
    assert (
        asyncio.run(
            registry["before_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
            ),
        )
        == "raw"
    )
    assert (
        asyncio.run(
            registry["wrap_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=ok_output,
            ),
        )
        == "raw"
    )
    assert asyncio.run(
        registry["after_output_validate"][0].func(
            cast("Any", None),
            output_context=_HOOK_CONTEXT,
            output={"ok": True},
        ),
    ) == {"ok": True}
    assert asyncio.run(
        registry["before_output_process"][0].func(
            cast("Any", None),
            output_context=_HOOK_CONTEXT,
            output={"ok": True},
        ),
    ) == {"ok": True}
    assert asyncio.run(
        registry["wrap_output_process"][0].func(
            cast("Any", None),
            output_context=_HOOK_CONTEXT,
            output={"ok": True},
            handler=ok_output,
        ),
    ) == {"ok": True}
    assert (
        asyncio.run(
            registry["after_output_process"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="done",
            ),
        )
        == "done"
    )
    assert (
        asyncio.run(
            registry["handle_deferred_tool_calls"][0].func(
                cast("Any", None),
                requests=cast("Any", None),
            ),
        )
        is None
    )

    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None
    titles = [update.title for update in updates if isinstance(update, ToolCallProgress)]
    assert "hook.prepare_output_tools" in titles
    assert "hook.wrap_output_validate" in titles
    assert "hook.wrap_output_process" in titles
    assert "hook.handle_deferred_tool_calls" in titles


def test_hook_bridge_output_error_hooks_emit_failed_updates() -> None:
    bridge = HookBridge()
    session = AcpSessionContext(
        session_id="hook-output-errors",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    registry = bridge.build_capability(session)._registry

    async def fail_output(output: Any) -> Any:
        del output
        raise RuntimeError("output failed")

    with pytest.raises(RuntimeError, match="output failed"):
        asyncio.run(
            registry["wrap_output_validate"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=fail_output,
            ),
        )
    with pytest.raises(RuntimeError, match="validate direct"):
        asyncio.run(
            registry["on_output_validate_error"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                error=RuntimeError("validate direct"),
            ),
        )
    with pytest.raises(RuntimeError, match="output failed"):
        asyncio.run(
            registry["wrap_output_process"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                handler=fail_output,
            ),
        )
    with pytest.raises(RuntimeError, match="process direct"):
        asyncio.run(
            registry["on_output_process_error"][0].func(
                cast("Any", None),
                output_context=_HOOK_CONTEXT,
                output="raw",
                error=RuntimeError("process direct"),
            ),
        )

    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None
    failed_titles = [
        update.title
        for update in updates
        if isinstance(update, ToolCallProgress) and update.status == "failed"
    ]
    assert failed_titles == [
        "hook.wrap_output_validate",
        "hook.on_output_validate_error",
        "hook.wrap_output_process",
        "hook.on_output_process_error",
    ]


def test_hook_bridge_skips_recording_when_flags_are_disabled_after_binding() -> None:
    bridge = HookBridge()
    session = AcpSessionContext(
        session_id="hook-disabled-paths",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    capability = bridge.build_capability(session)
    registry = capability._registry
    tool_call = ToolCallPart("echo", {"text": "hello"})
    tool_def = ToolDefinition(name="echo")
    request_context = cast("Any", SimpleNamespace(messages=[1]))
    response = ModelResponse(parts=[])

    bridge.record_run_lifecycle = False
    bridge.record_node_lifecycle = False
    bridge.record_event_stream = False
    bridge.record_model_requests = False
    bridge.record_prepare_tools = False
    bridge.record_tool_validation = False
    bridge.record_tool_execution = False

    async def ok_run() -> Any:
        return cast("Any", "run-result")

    async def fail_run() -> Any:
        raise RuntimeError("run failed")

    async def ok_node(node: Any) -> Any:
        return node

    async def fail_node(node: Any) -> Any:
        del node
        raise RuntimeError("node failed")

    async def ok_model(request_context: Any) -> ModelResponse:
        del request_context
        return response

    async def fail_model(request_context: Any) -> ModelResponse:
        del request_context
        raise RuntimeError("model failed")

    async def ok_validate(args: Any) -> Any:
        return args

    async def fail_validate(args: Any) -> Any:
        del args
        raise RuntimeError("validate failed")

    async def ok_execute(args: Any) -> Any:
        return {"ok": args["text"]}

    async def fail_execute(args: Any) -> Any:
        del args
        raise RuntimeError("execute failed")

    async def event_stream():
        yield cast("Any", SimpleNamespace(event_kind="demo_event"))

    assert asyncio.run(registry["before_run"][0].func(cast("Any", None))) is None
    assert (
        asyncio.run(registry["wrap_run"][0].func(cast("Any", None), handler=ok_run)) == "run-result"
    )
    with pytest.raises(RuntimeError, match="run failed"):
        asyncio.run(registry["wrap_run"][0].func(cast("Any", None), handler=fail_run))
    assert asyncio.run(registry["after_run"][0].func(cast("Any", None), result="done")) == "done"
    with pytest.raises(RuntimeError, match="run direct"):
        asyncio.run(
            registry["on_run_error"][0].func(cast("Any", None), error=RuntimeError("run direct")),
        )
    assert (
        asyncio.run(registry["before_node_run"][0].func(cast("Any", None), node=tool_call))
        is tool_call
    )
    assert (
        asyncio.run(
            registry["after_node_run"][0].func(
                cast("Any", None),
                node=tool_call,
                result="node-result",
            ),
        )
        == "node-result"
    )
    assert (
        asyncio.run(
            registry["wrap_node_run"][0].func(cast("Any", None), node=tool_call, handler=ok_node),
        )
        is tool_call
    )
    with pytest.raises(RuntimeError, match="node failed"):
        asyncio.run(
            registry["wrap_node_run"][0].func(cast("Any", None), node=tool_call, handler=fail_node),
        )
    assert (
        asyncio.run(registry["_on_event"][0].func(cast("Any", None), cast("Any", tool_call)))
        is tool_call
    )
    wrapped_stream = registry["wrap_run_event_stream"][0].func(
        cast("Any", None),
        stream=event_stream(),
    )
    assert asyncio.run(wrapped_stream.__anext__()).event_kind == "demo_event"
    assert (
        asyncio.run(registry["before_model_request"][0].func(cast("Any", None), request_context))
        is request_context
    )
    assert (
        asyncio.run(
            registry["wrap_model_request"][0].func(
                cast("Any", None),
                request_context=request_context,
                handler=ok_model,
            ),
        )
        is response
    )
    with pytest.raises(RuntimeError, match="model failed"):
        asyncio.run(
            registry["wrap_model_request"][0].func(
                cast("Any", None),
                request_context=request_context,
                handler=fail_model,
            ),
        )
    assert (
        asyncio.run(
            registry["after_model_request"][0].func(
                cast("Any", None),
                request_context=request_context,
                response=response,
            ),
        )
        is response
    )
    assert asyncio.run(registry["prepare_tools"][0].func(cast("Any", None), [tool_def])) == [
        tool_def,
    ]
    assert asyncio.run(
        registry["before_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["after_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["wrap_tool_validate"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
            handler=ok_validate,
        ),
    ) == {"text": "hello"}
    with pytest.raises(RuntimeError, match="validate failed"):
        asyncio.run(
            registry["wrap_tool_validate"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                handler=fail_validate,
            ),
        )
    assert asyncio.run(
        registry["before_tool_execute"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
        ),
    ) == {"text": "hello"}
    assert asyncio.run(
        registry["wrap_tool_execute"][0].func(
            cast("Any", None),
            call=tool_call,
            tool_def=tool_def,
            args={"text": "hello"},
            handler=ok_execute,
        ),
    ) == {"ok": "hello"}
    with pytest.raises(RuntimeError, match="execute failed"):
        asyncio.run(
            registry["wrap_tool_execute"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                handler=fail_execute,
            ),
        )
    assert (
        asyncio.run(
            registry["after_tool_execute"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                result="done",
            ),
        )
        == "done"
    )
    with pytest.raises(RuntimeError, match="execute direct"):
        asyncio.run(
            registry["on_tool_execute_error"][0].func(
                cast("Any", None),
                call=tool_call,
                tool_def=tool_def,
                args={"text": "hello"},
                error=RuntimeError("execute direct"),
            ),
        )

    assert bridge.drain_updates(session, Agent(TestModel(custom_output_text="unused"))) is None


def test_external_hook_event_bridge_records_modes_and_metadata() -> None:
    session = AcpSessionContext(
        session_id="external-hooks",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    bridge = ExternalHookEventBridge()
    event = HookEvent(
        event_id="before_run",
        hook_name="before_run",
        tool_name=None,
        tool_filters=(),
        raw_output="done",
        status="completed",
    )

    bridge.record_event(session, event)
    metadata = bridge.get_session_metadata(session, Agent(TestModel()))
    updates = bridge.drain_updates(session, Agent(TestModel()))

    assert metadata["pending_event_count"] == 2
    assert updates is not None
    assert [update.tool_call_id for update in updates] == [
        "external-hooks:external-hook:1",
        "external-hooks:external-hook:1",
    ]
    assert isinstance(updates[0], ToolCallStart)
    assert isinstance(updates[1], ToolCallProgress)
    assert updates[1].status == "completed"
    assert bridge.drain_updates(session, Agent(TestModel())) is None


def test_external_hook_event_bridge_start_only_and_hidden_events() -> None:
    session = AcpSessionContext(
        session_id="external-hooks-hidden",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    bridge = ExternalHookEventBridge(
        projection_map=HookProjectionMap(hidden_event_ids=frozenset({"secret"})),
        emission_mode="start_only",
    )

    bridge.record_event(
        session,
        HookEvent(
            event_id="secret",
            hook_name="secret",
            tool_name=None,
            tool_filters=(),
            status="completed",
        ),
    )
    bridge.record_event(
        session,
        HookEvent(
            event_id="before_run",
            hook_name="before_run",
            tool_name=None,
            tool_filters=(),
            status="failed",
        ),
    )

    updates = bridge.drain_updates(session, Agent(TestModel()))

    assert updates is not None
    assert len(updates) == 1
    assert isinstance(updates[0], ToolCallStart)
    assert updates[0].status == "failed"


def test_external_hook_event_bridge_start_only_preserves_in_progress_without_status() -> None:
    session = AcpSessionContext(
        session_id="external-hooks-start-only-default-status",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    bridge = ExternalHookEventBridge(emission_mode="start_only")

    bridge.record_event(
        session,
        HookEvent(
            event_id="before_run",
            hook_name="before_run",
            tool_name=None,
            tool_filters=(),
            status=None,
        ),
    )

    updates = bridge.drain_updates(session, Agent(TestModel()))

    assert updates is not None
    assert len(updates) == 1
    assert isinstance(updates[0], ToolCallStart)
    assert updates[0].status == "in_progress"


def test_external_hook_event_bridge_keeps_start_when_progress_is_hidden() -> None:
    session = AcpSessionContext(
        session_id="external-hooks-start-only-progress-hidden",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    bridge = ExternalHookEventBridge()

    bridge.record_event(
        session,
        HookEvent(
            event_id="before_run",
            hook_name="before_run",
            tool_name=None,
            tool_filters=(),
            status=None,
        ),
    )

    updates = bridge.drain_updates(session, Agent(TestModel()))

    assert updates is not None
    assert len(updates) == 1
    assert isinstance(updates[0], ToolCallStart)
    assert updates[0].status == "in_progress"



from __future__ import annotations as _annotations

import os
from collections.abc import Callable, Sequence
from importlib.util import find_spec
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from acp.schema import ModelInfo, SessionMode
from codex_auth_helper import create_codex_chat_openai
from langchain_acp import (
    AcpSessionContext,
    AdapterConfig,
    DeepAgentsCompatibilityBridge,
    DeepAgentsProjectionMap,
    FileSessionStore,
    create_acp_agent,
    native_plan_tools,
    run_acp,
)
from langchain_acp.session import utc_now

if TYPE_CHECKING:
    from langchain_acp import CompiledAgentGraph
    from langchain_core.tools import BaseTool

DeepAgentTool: TypeAlias = "BaseTool | Callable[..., Any] | dict[str, Any]"

__all__ = (
    "AVAILABLE_MODELS",
    "AVAILABLE_MODES",
    "MODEL_NAME",
    "WORKSPACE_ROOT",
    "acp_agent",
    "codex_instructions",
    "config",
    "graph",
    "graph_from_session",
    "list_workspace_files",
    "main",
    "read_file",
    "write_file",
)

_DEMO_ROOT = Path("agent_demos")
WORKSPACE_ROOT = Path.cwd() / _DEMO_ROOT / "deepagents-graph"
MODEL_NAME = os.getenv("CODEX_MODEL", "gpt-5.4")
_SESSION_ROOT = _DEMO_ROOT / "deepagents-graph"
_SESSION_STORE_ROOT = (
    Path(os.getenv("ACP_EXAMPLE_SESSION_DIR", str(_DEMO_ROOT / "acp-sessions")))
    .expanduser()
    .resolve()
    / "langchain-deepagents"
)
MOCK_WORKSPACE_FILES = {
    "brief.md": (
        "# DeepAgents Demo\n\nThis is a mocked workspace file used by the DeepAgents example.\n"
    ),
    "notes.md": (
        "# Mock Notes\n\n"
        "- This example uses fixed filesystem tool outputs.\n"
        "- No real workspace mutation happens here.\n"
    ),
}
AVAILABLE_MODELS = (
    ModelInfo(model_id="gpt-5.4-mini", name="GPT-5.4 Mini"),
    ModelInfo(model_id=MODEL_NAME, name="GPT-5.4"),
)
AVAILABLE_MODES = (
    SessionMode(id="ask", name="Ask", description="Inspect the workspace and answer questions."),
    SessionMode(id="edit", name="Edit", description="Make direct workspace changes."),
    SessionMode(id="plan", name="Plan", description="Plan before applying changes."),
)
DEFAULT_MODEL_ID = AVAILABLE_MODELS[0].model_id
DEFAULT_MODE_ID = AVAILABLE_MODES[0].id


def codex_instructions(*, mode_id: str) -> str:
    """Return the Codex Responses instructions used by the DeepAgents example."""
    base = (
        "You are a workspace agent that can inspect and update files in the current demo workspace. "
        "Use tools for concrete file work, keep changes scoped, and summarize what changed."
    )
    if mode_id == "edit":
        return f"{base} Prefer direct file edits when the user asks for changes."
    if mode_id == "plan":
        return f"{base} Begin with a short plan before making any edits."
    return base


def _deepagents_available() -> bool:
    return find_spec("deepagents") is not None


def _ensure_workspace(root: Path | None = None) -> Path:
    if root is None:
        root = WORKSPACE_ROOT
    root.mkdir(parents=True, exist_ok=True)
    brief_path = root / "brief.md"
    if not brief_path.exists():
        brief_path.write_text(
            "# DeepAgents Demo\n\n"
            "This seeded file exists so ACP can render read and write projections.\n",
            encoding="utf-8",
        )
    return root


def _resolve_workspace_path(path: str, *, root: Path | None = None) -> Path:
    if root is None:
        root = WORKSPACE_ROOT
    workspace_root = _ensure_workspace(root).resolve()
    workspace_parent = workspace_root.parent
    requested_path = Path(path)
    if requested_path.is_absolute():
        absolute_candidate = requested_path.resolve()
        try:
            relative_to_workspace = absolute_candidate.relative_to(workspace_root)
        except ValueError:
            try:
                relative_to_parent = absolute_candidate.relative_to(workspace_parent)
            except ValueError:
                candidate = absolute_candidate
            else:
                relative_to_parent = Path(
                    *(
                        relative_to_parent.parts[1:]
                        if relative_to_parent.parts
                        and relative_to_parent.parts[0] == workspace_root.name
                        else relative_to_parent.parts
                    ),
                )
                candidate = (workspace_root / relative_to_parent).resolve()
        else:
            candidate = (workspace_root / relative_to_workspace).resolve()
    else:
        relative_path = requested_path
        if requested_path.parts and requested_path.parts[0] == workspace_root.name:
            relative_path = Path(*requested_path.parts[1:])
        candidate = (workspace_root / relative_path).resolve()
    try:
        candidate.relative_to(workspace_root)
    except ValueError as exc:
        raise ValueError("Path must stay inside the DeepAgents example workspace.") from exc
    return candidate


def _normalized_mock_path(path: str) -> str:
    normalized = Path(path).name.strip()
    if not normalized:
        raise ValueError("Path must reference a mocked workspace file.")
    return normalized


def list_workspace_files() -> str:
    """List mocked workspace files exposed by the DeepAgents example."""
    return "\n".join(sorted(MOCK_WORKSPACE_FILES))


def read_file(path: str) -> str:
    """Read a mocked file from the DeepAgents example workspace."""
    normalized = _normalized_mock_path(path)
    content = MOCK_WORKSPACE_FILES.get(normalized)
    if content is None:
        raise ValueError(f"File not found: {path}")
    return content


def write_file(path: str, content: str) -> str:
    """Pretend to write a file in the DeepAgents example workspace."""
    normalized = _normalized_mock_path(path)
    del content
    return f"Mock write accepted for {normalized}"


def _session_workspace_root(session: AcpSessionContext) -> Path:
    return session.cwd.resolve() / _SESSION_ROOT


def _bind_workspace_tools(
    root: Path,
) -> tuple[Callable[[], str], Callable[[str], str], Callable[[str, str], str]]:
    del root

    def _list_workspace_files() -> str:
        return "\n".join(sorted(MOCK_WORKSPACE_FILES))

    _list_workspace_files.__name__ = "list_workspace_files"
    _list_workspace_files.__doc__ = list_workspace_files.__doc__

    def _read_file(path: str) -> str:
        normalized = _normalized_mock_path(path)
        content = MOCK_WORKSPACE_FILES.get(normalized)
        if content is None:
            raise ValueError(f"File not found: {path}")
        return content

    _read_file.__name__ = "read_file"
    _read_file.__doc__ = read_file.__doc__

    def _write_file(path: str, content: str) -> str:
        del content
        normalized = _normalized_mock_path(path)
        return f"Mock write accepted for {normalized}"

    _write_file.__name__ = "write_file"
    _write_file.__doc__ = write_file.__doc__
    return (_list_workspace_files, _read_file, _write_file)


def graph_from_session(session: AcpSessionContext) -> CompiledAgentGraph:
    if not _deepagents_available():
        raise RuntimeError(
            'Install the optional DeepAgents dependency first: uv add "langchain-acp[deepagents]"',
        )
    workspace_root = _ensure_workspace(_session_workspace_root(session)).resolve()
    model_name = session.session_model_id or DEFAULT_MODEL_ID or MODEL_NAME
    mode_id = session.session_mode_id or DEFAULT_MODE_ID or "ask"
    from deepagents import create_deep_agent

    tools: Sequence[DeepAgentTool] = [
        *_bind_workspace_tools(workspace_root),
        *cast("Sequence[DeepAgentTool]", native_plan_tools()),
    ]
    return create_deep_agent(
        model=create_codex_chat_openai(
            model_name,
            instructions=codex_instructions(mode_id=mode_id),
        ),
        tools=tools,
        interrupt_on={"write_file": True},
        name=f"deepagents-{mode_id}-{session.cwd.name}",
    )


def _seed_session() -> AcpSessionContext:
    root = _ensure_workspace().resolve()
    timestamp = utc_now()
    return AcpSessionContext(
        session_id="deepagents-example",
        cwd=root,
        created_at=timestamp,
        updated_at=timestamp,
    )


graph = graph_from_session(_seed_session()) if _deepagents_available() else None

config = AdapterConfig(
    available_models=list(AVAILABLE_MODELS),
    available_modes=list(AVAILABLE_MODES),
    session_store=FileSessionStore(_SESSION_STORE_ROOT),
    capability_bridges=[DeepAgentsCompatibilityBridge()],
    default_model_id=DEFAULT_MODEL_ID,
    default_mode_id=DEFAULT_MODE_ID,
    default_plan_generation_type="tools",
    enable_plan_progress_tools=True,
    plan_mode_id="plan",
    projection_maps=[DeepAgentsProjectionMap()],
)
acp_agent = create_acp_agent(graph_factory=graph_from_session, config=config)


def main() -> None:
    _ensure_workspace()
    run_acp(graph_factory=graph_from_session, config=config)


if __name__ == "__main__":
    main()



from __future__ import annotations as _annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from acp.exceptions import RequestError
from pydantic_ai import Agent as PydanticAgent

from ..models import AdapterModel, ModelOverride
from ..session.state import AcpSessionContext
from ._agent_state import (
    assign_model,
    clear_selected_model_id,
    default_model,
    remember_default_model,
    selected_model_id,
    set_selected_model_id,
)
from .slash_commands import (
    extract_session_mcp_servers,
    list_agent_tools,
    render_hook_listing,
    render_mcp_server_listing,
    render_mode_message,
    render_model_message,
    render_thinking_message,
    render_tool_listing,
)

if TYPE_CHECKING:
    from ._session_runtime import _SessionRuntime

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

__all__ = ("_SessionModelRuntime",)


class _SessionModelRuntime(Generic[AgentDepsT, OutputDataT]):
    def __init__(self, runtime: _SessionRuntime[AgentDepsT, OutputDataT]) -> None:
        self._runtime = runtime

    async def handle_slash_command(
        self,
        command_name: str,
        *,
        argument: str | None,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> str | None:
        mode_state = await self._runtime._get_mode_state(session, agent)
        if mode_state is not None and command_name in {mode.id for mode in mode_state.modes}:
            if await self._runtime.set_session_mode(command_name, session.session_id) is None:
                return f"Mode `{command_name}` is unavailable"
            return render_mode_message(command_name)
        if command_name == "model":
            if argument is None:
                return render_model_message(self._runtime._resolve_current_model_id(session, agent))
            self._runtime._set_agent_model(agent, session, argument)
            surface = await self._runtime._build_session_surface(session, agent)
            await self._runtime._emit_session_state_updates(
                session,
                surface,
                emit_available_commands=True,
                emit_config_options=True,
                emit_current_mode=False,
                emit_plan=True,
                emit_session_info=True,
            )
            return render_model_message(self._runtime._resolve_current_model_id(session, agent))
        if command_name == "thinking":
            if argument is None:
                return render_thinking_message(
                    self._runtime._current_thinking_value(session, agent),
                )
            normalized_argument = argument.strip().lower()
            if (
                await self._runtime.set_config_option(
                    "thinking",
                    session.session_id,
                    normalized_argument,
                )
                is None
            ):
                return "Thinking effort is unavailable or invalid"
            return render_thinking_message(normalized_argument)
        if command_name == "tools":
            return render_tool_listing(list_agent_tools(agent))
        if command_name == "hooks":
            return render_hook_listing(
                self._runtime._owner._list_agent_hooks(agent),
                projection_map=self._runtime._owner._config.hook_projection_map,
            )
        if command_name == "mcp-servers":
            return render_mcp_server_listing(
                extract_session_mcp_servers(session, agent=agent),
            )
        return None

    def apply_session_model_to_agent(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
    ) -> None:
        session_model_id = session.session_model_id
        if session_model_id is None:
            if selected_model_id(agent) is not None:
                self.restore_agent_default_model(agent)
            return
        if selected_model_id(agent) == session_model_id:
            return
        self.set_agent_model(agent, session, session_model_id)

    def set_agent_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
        model_id: str,
    ) -> None:
        self.remember_default_model(agent)
        selected_model = self.resolve_selected_model(model_id)
        assign_model(agent, selected_model)
        normalized_model_id = model_id.strip()
        set_selected_model_id(agent, normalized_model_id)
        session.session_model_id = normalized_model_id
        session.config_values["model"] = normalized_model_id

    def restore_default_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
    ) -> bool:
        if session.session_model_id is None:
            return False
        if not self.restore_agent_default_model(agent):
            return False
        restored_model_id = self._runtime._resolve_model_id_from_value(default_model(agent))
        session.session_model_id = restored_model_id
        if restored_model_id is None:
            session.config_values.pop("model", None)
        else:
            session.config_values["model"] = restored_model_id
        return True

    def restore_agent_default_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> bool:
        agent_default_model = default_model(agent)
        if agent_default_model is None:
            return False
        assign_model(agent, agent_default_model)
        clear_selected_model_id(agent)
        return True

    def remember_default_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        remember_default_model(agent)

    def resolve_selected_model(self, model_id: str) -> ModelOverride:
        normalized_model_id = model_id.strip()
        if not normalized_model_id:
            raise RequestError.invalid_params({"modelId": model_id})
        model_option = self._runtime._find_model_option(normalized_model_id)
        if model_option is not None:
            return self.resolve_model_option(model_option)
        return self.resolve_unconfigured_model_id(normalized_model_id)

    def resolve_model_option(self, model_option: AdapterModel) -> ModelOverride:
        if (
            isinstance(model_option.override, str)
            and model_option.override.strip() == model_option.model_id.strip()
        ):
            return self.resolve_unconfigured_model_id(model_option.model_id)
        return model_option.override

    def resolve_unconfigured_model_id(self, model_id: str) -> ModelOverride:
        normalized_model_id = model_id.strip()
        if normalized_model_id.startswith("codex:"):
            codex_model_id = normalized_model_id.removeprefix("codex:").strip()
            if not codex_model_id:
                raise RequestError.invalid_params({"modelId": model_id})
            try:
                from codex_auth_helper import create_codex_responses_model
            except ImportError as exc:
                raise RequestError.invalid_params({"modelId": normalized_model_id}) from exc
            return create_codex_responses_model(
                codex_model_id,
                instructions=(
                    "Follow the agent instructions already present in the request context "
                    "and answer consistently with the selected Codex model."
                ),
            )
        return normalized_model_id



from __future__ import annotations as _annotations

import asyncio
import builtins
import os
import sys
import threading
from concurrent.futures import ThreadPoolExecutor
from types import ModuleType, SimpleNamespace
from typing import Any, cast

import pytest
from acp.schema import ToolKind
from pydantic_acp import (
    HarnessCodeModeBridge,
    HarnessFileSystemBridge,
    HarnessShellBridge,
)
from pydantic_acp.bridges.capability_support import (
    _json_user_location,
    _resolve_mcp_server_id,
)
from pydantic_acp.projection import (
    HarnessCodeModeProjectionMap,
    HarnessFileSystemProjectionMap,
    HarnessShellProjectionMap,
)
from pydantic_acp.session.state import utc_now
from pydantic_ai import ModelRequestContext, ModelResponse
from pydantic_ai.capabilities import MCP, ImageGeneration, Toolset
from pydantic_ai.messages import (
    CompactionPart,
    ModelMessage,
    ModelRequest,
    TextPart,
    UserPromptPart,
)
from pydantic_ai.models.function import AgentInfo, FunctionModel
from pydantic_ai.models.openai import OpenAIResponsesModel
from pydantic_ai.toolsets.function import FunctionToolset

from .support import (
    AcpSessionContext,
    AdapterConfig,
    Agent,
    AgentBridgeBuilder,
    AnthropicCompactionBridge,
    ImageGenerationBridge,
    IncludeToolReturnSchemasBridge,
    McpCapabilityBridge,
    MemorySessionStore,
    OpenAICompactionBridge,
    Path,
    PrefixToolsBridge,
    RecordingClient,
    SetToolMetadataBridge,
    TestModel,
    ThreadExecutorBridge,
    ToolCallProgress,
    ToolCallStart,
    ToolsetBridge,
    create_acp_agent,
    text_block,
)


class _FakeHarnessCapability:
    def __init__(self, **kwargs: Any) -> None:
        self.kwargs = kwargs


def _install_fake_harness_modules(monkeypatch: pytest.MonkeyPatch) -> None:
    harness_module = ModuleType("pydantic_ai_harness")
    filesystem_module = ModuleType("pydantic_ai_harness.filesystem")
    shell_module = ModuleType("pydantic_ai_harness.shell")
    code_mode_module = ModuleType("pydantic_ai_harness.code_mode")

    cast("Any", filesystem_module).FileSystem = _FakeHarnessCapability
    cast("Any", shell_module).Shell = _FakeHarnessCapability
    cast("Any", code_mode_module).CodeMode = _FakeHarnessCapability

    monkeypatch.setitem(sys.modules, "pydantic_ai_harness", harness_module)
    monkeypatch.setitem(sys.modules, "pydantic_ai_harness.filesystem", filesystem_module)
    monkeypatch.setitem(sys.modules, "pydantic_ai_harness.shell", shell_module)
    monkeypatch.setitem(sys.modules, "pydantic_ai_harness.code_mode", code_mode_module)


def _test_session() -> AcpSessionContext:
    now = utc_now()
    return AcpSessionContext(
        session_id="s",
        cwd=Path("/workspace"),
        created_at=now,
        updated_at=now,
    )


def _write_mcp_stdio_server_script(path: Path) -> None:
    path.write_text(
        "\n".join(
            (
                "from __future__ import annotations as _annotations",
                "",
                "from mcp.server.fastmcp import FastMCP",
                "",
                'mcp = FastMCP("test-mcp", instructions="Be a helpful assistant.")',
                "",
                "@mcp.tool()",
                "def ping() -> str:",
                '    return "pong"',
                "",
                'if __name__ == "__main__":',
                '    mcp.run("stdio")',
                "",
            ),
        ),
        encoding="utf-8",
    )


def _build_mcp_stdio_test_env(
    *,
    executable: str,
    base_executable: str | None,
    sys_path: list[str],
    environ: dict[str, str],
) -> tuple[str, dict[str, str]]:
    executable_path = Path(executable)
    python_executable = (
        str(executable_path) if executable_path.exists() else base_executable or executable
    )
    python_path_entries = [entry for entry in sys_path if entry]
    mcp_env = dict(environ)
    if python_path_entries:
        existing_pythonpath = mcp_env.get("PYTHONPATH")
        combined_pythonpath = os.pathsep.join(
            (
                *python_path_entries,
                *([existing_pythonpath] if existing_pythonpath else []),
            ),
        )
        mcp_env["PYTHONPATH"] = combined_pythonpath
    return python_executable, mcp_env


def test_harness_filesystem_bridge_builds_capability_and_metadata(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_harness_modules(monkeypatch)
    session = _test_session()
    bridge = HarnessFileSystemBridge(
        root_dir=Path("/workspace"),
        allowed_patterns=("src/**",),
        denied_patterns=("*.secret",),
        protected_patterns=(".git/*", ".env"),
        max_read_lines=50,
        max_search_results=12,
        max_find_results=7,
    )

    capability = bridge.build_capability(session)
    capability_tuple = bridge.build_agent_capabilities(session)
    assert isinstance(capability, _FakeHarnessCapability)
    assert capability_tuple is not None
    assert isinstance(capability_tuple[0], _FakeHarnessCapability)
    assert capability.kwargs == {
        "root_dir": Path("/workspace"),
        "allowed_patterns": ["src/**"],
        "denied_patterns": ["*.secret"],
        "max_read_lines": 50,
        "max_search_results": 12,
        "max_find_results": 7,
        "protected_patterns": [".git/*", ".env"],
    }
    assert isinstance(bridge.get_projection_maps()[0], HarnessFileSystemProjectionMap)
    assert bridge.get_session_metadata(session, cast("Any", object())) == {
        "allowed_patterns": ["src/**"],
        "denied_patterns": ["*.secret"],
        "max_find_results": 7,
        "max_read_lines": 50,
        "max_search_results": 12,
        "protected_patterns": [".env", ".git/*"],
        "root_dir": "/workspace",
        "tool_names": [
            "create_directory",
            "edit_file",
            "list_directory",
            "read_file",
            "search_files",
            "write_file",
        ],
    }
    assert bridge.get_tool_kind("read_file") == "read"
    assert bridge.get_tool_kind("write_file") == "edit"
    assert bridge.get_tool_kind("edit_file") == "edit"
    assert bridge.get_tool_kind("search_files") == "search"
    assert bridge.get_tool_kind("other") is None


def test_harness_shell_bridge_builds_capability_and_metadata(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_harness_modules(monkeypatch)
    session = _test_session()
    bridge = HarnessShellBridge(
        cwd=Path("/workspace"),
        allowed_commands=("git", "pytest"),
        denied_commands=("rm",),
        denied_operators=(">",),
        default_timeout=2.5,
        max_output_chars=123,
        persist_cwd=True,
        allow_interactive=True,
        env={"SAFE": "1"},
        denied_env_patterns=("OPENAI_*",),
    )

    capability = bridge.build_capability(session)
    capability_tuple = bridge.build_agent_capabilities(session)
    assert isinstance(capability, _FakeHarnessCapability)
    assert capability_tuple is not None
    assert isinstance(capability_tuple[0], _FakeHarnessCapability)
    assert capability.kwargs == {
        "cwd": Path("/workspace"),
        "allowed_commands": ["git", "pytest"],
        "denied_operators": [">"],
        "default_timeout": 2.5,
        "max_output_chars": 123,
        "persist_cwd": True,
        "allow_interactive": True,
        "env": {"SAFE": "1"},
        "denied_env_patterns": ["OPENAI_*"],
        "denied_commands": ["rm"],
    }
    assert isinstance(bridge.get_projection_maps()[0], HarnessShellProjectionMap)
    assert bridge.get_session_metadata(session, cast("Any", object())) == {
        "allow_interactive": True,
        "allowed_commands": ["git", "pytest"],
        "cwd": "/workspace",
        "default_timeout": 2.5,
        "denied_commands": ["rm"],
        "denied_env_patterns": ["OPENAI_*"],
        "denied_operators": [">"],
        "env_keys": ["SAFE"],
        "max_output_chars": 123,
        "persist_cwd": True,
        "tool_names": ["check_command", "run_command", "start_command", "stop_command"],
    }
    assert bridge.get_tool_kind("run_command") == "execute"
    assert bridge.get_tool_kind("not_shell") is None


def test_harness_bridges_omit_none_optional_constructor_kwargs(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_harness_modules(monkeypatch)
    session = _test_session()

    filesystem_capability = HarnessFileSystemBridge().build_capability(session)
    shell_capability = HarnessShellBridge().build_capability(session)

    assert isinstance(filesystem_capability, _FakeHarnessCapability)
    assert isinstance(shell_capability, _FakeHarnessCapability)
    assert "protected_patterns" not in filesystem_capability.kwargs
    assert "denied_commands" not in shell_capability.kwargs


def test_harness_code_mode_bridge_builds_capability_and_metadata(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_harness_modules(monkeypatch)
    session = _test_session()
    os_access = object()
    mount = object()
    bridge = HarnessCodeModeBridge(
        max_retries=5,
        os_access=os_access,
        mount=mount,
        dynamic_catalog=True,
    )

    capability = bridge.build_capability(session)
    capability_tuple = bridge.build_agent_capabilities(session)
    assert isinstance(capability, _FakeHarnessCapability)
    assert capability_tuple is not None
    assert isinstance(capability_tuple[0], _FakeHarnessCapability)
    assert capability.kwargs == {
        "tools": "all",
        "max_retries": 5,
        "os_access": os_access,
        "mount": mount,
        "dynamic_catalog": True,
    }
    assert isinstance(bridge.get_projection_maps()[0], HarnessCodeModeProjectionMap)
    assert bridge.get_session_metadata(session, cast("Any", object())) == {
        "dynamic_catalog": True,
        "has_mount": True,
        "has_os_access": True,
        "max_retries": 5,
        "tool_names": ["run_code"],
    }
    assert bridge.get_tool_kind("run_code") == "execute"
    assert bridge.get_tool_kind("other") is None


def test_harness_v07_bridges_build_real_capabilities() -> None:
    from pydantic_ai_harness.code_mode import CodeMode
    from pydantic_ai_harness.filesystem import FileSystem
    from pydantic_ai_harness.shell import Shell

    session = _test_session()

    assert isinstance(
        HarnessFileSystemBridge(root_dir=Path(".")).build_capability(session),
        FileSystem,
    )
    assert isinstance(
        HarnessShellBridge(cwd=Path(".")).build_capability(session),
        Shell,
    )
    assert isinstance(HarnessCodeModeBridge().build_capability(session), CodeMode)


def test_harness_bridges_report_missing_optional_dependency(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setitem(sys.modules, "pydantic_ai_harness", ModuleType("pydantic_ai_harness"))
    for module_name in tuple(sys.modules):
        if module_name.startswith("pydantic_ai_harness"):
            monkeypatch.delitem(sys.modules, module_name, raising=False)
    real_import = builtins.__import__

    def missing_harness_import(
        name: str,
        globals: dict[str, Any] | None = None,
        locals: dict[str, Any] | None = None,
        fromlist: tuple[str, ...] = (),
        level: int = 0,
    ) -> Any:
        if name.startswith("pydantic_ai_harness"):
            raise ImportError(name)
        return real_import(name, globals, locals, fromlist, level)

    assert missing_harness_import("math").__name__ == "math"
    monkeypatch.setattr(builtins, "__import__", missing_harness_import)
    session = _test_session()

    with pytest.raises(ImportError, match="pydantic-ai-harness"):
        HarnessFileSystemBridge().build_capability(session)
    with pytest.raises(ImportError, match="pydantic-ai-harness"):
        HarnessShellBridge().build_capability(session)
    with pytest.raises(ImportError, match=r"pydantic-ai-harness\[code-mode\]"):
        HarnessCodeModeBridge().build_capability(session)


def test_harness_bridge_projection_maps_are_added_to_adapter_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_harness_modules(monkeypatch)
    adapter = create_acp_agent(
        agent=Agent(TestModel()),
        config=AdapterConfig(
            capability_bridges=[
                HarnessFileSystemBridge(),
                HarnessShellBridge(),
                HarnessCodeModeBridge(),
            ],
            session_store=MemorySessionStore(),
        ),
    )

    config = cast("Any", adapter)._config
    projection_map_types = {type(projection_map) for projection_map in config.projection_maps}
    assert HarnessFileSystemProjectionMap in projection_map_types
    assert HarnessShellProjectionMap in projection_map_types
    assert HarnessCodeModeProjectionMap in projection_map_types

    classifier = cast("Any", adapter)._tool_classifier
    assert cast("ToolKind", classifier.classify("read_file")) == "read"
    assert cast("ToolKind", classifier.classify("run_command")) == "execute"
    assert cast("ToolKind", classifier.classify("run_code")) == "execute"


def test_thread_executor_bridge_runs_sync_tools_on_configured_executor(
    tmp_path: Path,
) -> None:
    executor = ThreadPoolExecutor(max_workers=1, thread_name_prefix="acpkit-bridge")
    bridge = ThreadExecutorBridge(executor=executor)

    try:

        def factory(session: AcpSessionContext) -> Agent[None, str]:
            builder = AgentBridgeBuilder(
                session=session,
                capability_bridges=[bridge],
            )
            contributions = builder.build()
            agent = Agent(
                TestModel(call_tools=["check_thread"], custom_output_text="done"),
                capabilities=contributions.capabilities,
            )

            @agent.tool_plain
            def check_thread() -> str:
                return threading.current_thread().name

            return agent

        adapter = create_acp_agent(
            agent_factory=factory,
            config=AdapterConfig(
                capability_bridges=[bridge],
                session_store=MemorySessionStore(),
            ),
        )
        client = RecordingClient()
        adapter.on_connect(client)

        session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
        response = asyncio.run(
            adapter.prompt(
                prompt=[text_block("Check the thread executor.")],
                session_id=session.session_id,
            ),
        )

        assert response.stop_reason == "end_turn"
        progress_update = next(
            update
            for _, update in client.updates
            if isinstance(update, ToolCallProgress) and update.title == "check_thread"
        )
        assert isinstance(progress_update.raw_output, str)
        assert progress_update.raw_output.startswith("acpkit-bridge")
    finally:
        executor.shutdown(wait=True)


def test_metadata_and_return_schema_bridges_modify_selected_tools(
    tmp_path: Path,
) -> None:
    test_model = TestModel(custom_output_text="done")
    bridges = [
        SetToolMetadataBridge(tools=["tool_a"], code_mode=True),
        IncludeToolReturnSchemasBridge(tools=["tool_a"]),
    ]

    def factory(session: AcpSessionContext) -> Agent[None, str]:
        builder = AgentBridgeBuilder(
            session=session,
            capability_bridges=bridges,
        )
        contributions = builder.build()
        agent = Agent(
            test_model,
            capabilities=contributions.capabilities,
        )

        @agent.tool_plain
        def tool_a(x: int) -> int:
            return x

        @agent.tool_plain
        def tool_b(x: str) -> str:
            return x

        return agent

    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(
            capability_bridges=bridges,
            session_store=MemorySessionStore(),
        ),
    )

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    response = asyncio.run(
        adapter.prompt(
            prompt=[text_block("Inspect the available tools.")],
            session_id=session.session_id,
        ),
    )

    assert response.stop_reason == "end_turn"
    params = test_model.last_model_request_parameters
    assert params is not None
    tool_a_def = next(tool_def for tool_def in params.function_tools if tool_def.name == "tool_a")
    tool_b_def = next(tool_def for tool_def in params.function_tools if tool_def.name == "tool_b")
    assert tool_a_def.metadata is not None
    assert tool_a_def.metadata["code_mode"] is True
    assert tool_a_def.include_return_schema is True
    assert "Return schema" in (tool_a_def.description or "")
    assert tool_b_def.metadata is None or "code_mode" not in tool_b_def.metadata
    assert tool_b_def.include_return_schema is not True
    assert "Return schema" not in (tool_b_def.description or "")


def test_set_tool_metadata_bridge_can_attach_metadata_to_non_schema_tool(
    tmp_path: Path,
) -> None:
    test_model = TestModel(custom_output_text="done")
    bridge = SetToolMetadataBridge(tools=["tool_b"], code_mode=False)

    def factory(session: AcpSessionContext) -> Agent[None, str]:
        builder = AgentBridgeBuilder(
            session=session,
            capability_bridges=[bridge],
        )
        contributions = builder.build()
        agent = Agent(
            test_model,
            capabilities=contributions.capabilities,
        )

        @agent.tool_plain
        def tool_b(x: str) -> str:
            return x

        return agent

    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(
            capability_bridges=[bridge],
            session_store=MemorySessionStore(),
        ),
    )

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    response = asyncio.run(
        adapter.prompt(
            prompt=[text_block("Inspect the available tools.")],
            session_id=session.session_id,
        ),
    )

    assert response.stop_reason == "end_turn"
    params = test_model.last_model_request_parameters
    assert params is not None
    tool_b_def = next(tool_def for tool_def in params.function_tools if tool_def.name == "tool_b")
    assert tool_b_def.metadata is not None
    assert "code_mode" in tool_b_def.metadata


def test_image_generation_and_mcp_capability_bridges_build_metadata_and_classification() -> None:
    session = AcpSessionContext(
        session_id="session-1",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    image_bridge = ImageGenerationBridge(
        fallback_model="openai-responses:gpt-5.4",
        quality="high",
        size="1024x1024",
        output_format="png",
    )
    mcp_bridge = McpCapabilityBridge(
        url="https://example.com/services/repo/sse",
        local=False,
        allowed_tools=["search", "read_file"],
        description="Repo MCP",
        authorization_token="secret",
    )

    image_capability = image_bridge.build_capability(session)
    mcp_capability = mcp_bridge.build_capability(session)

    assert isinstance(image_capability, ImageGeneration)
    assert isinstance(mcp_capability, MCP)
    assert image_bridge.get_tool_kind("image_generation") == "execute"
    assert image_bridge.get_tool_kind("generate_image") == "execute"
    assert image_bridge.get_session_metadata(session, Agent(TestModel())) == {
        "aspect_ratio": None,
        "background": None,
        "fallback_model": "openai-responses:gpt-5.4",
        "input_fidelity": None,
        "moderation": None,
        "output_compression": None,
        "output_format": "png",
        "quality": "high",
        "size": "1024x1024",
        "tool_names": ["generate_image", "image_generation"],
    }
    mcp_metadata = mcp_bridge.get_session_metadata(session, Agent(TestModel()))
    assert mcp_bridge.get_tool_kind("mcp_server:repo") == "execute"
    assert mcp_metadata["allowed_tools"] == ["read_file", "search"]
    assert mcp_metadata["description"] == "Repo MCP"
    assert mcp_metadata["has_authorization_token"] is True
    assert mcp_metadata["headers"] == []
    assert mcp_metadata["server_id"] == "example.com-sse"
    assert mcp_metadata["url"] == "https://example.com/services/repo/sse"


def test_toolset_and_prefix_bridges_expose_function_tools_to_the_model(
    tmp_path: Path,
) -> None:
    toolset = FunctionToolset()

    @toolset.tool_plain
    def lookup(query: str) -> str:
        return f"lookup:{query}"

    prefixed_toolset = FunctionToolset()

    @prefixed_toolset.tool_plain
    def search(term: str) -> str:
        return f"search:{term}"

    plain_model = TestModel(custom_output_text="done")
    prefixed_model = TestModel(custom_output_text="done")

    def plain_factory(session: AcpSessionContext) -> Agent[None, str]:
        builder = AgentBridgeBuilder(
            session=session,
            capability_bridges=[ToolsetBridge(toolset=toolset)],
        )
        contributions = builder.build()
        return Agent(
            plain_model,
            capabilities=contributions.capabilities,
        )

    def prefixed_factory(session: AcpSessionContext) -> Agent[None, str]:
        builder = AgentBridgeBuilder(
            session=session,
            capability_bridges=[
                PrefixToolsBridge(
                    wrapped=Toolset(toolset=prefixed_toolset),
                    prefix="repo",
                ),
            ],
        )
        contributions = builder.build()
        return Agent(
            prefixed_model,
            capabilities=contributions.capabilities,
        )

    plain_adapter = create_acp_agent(
        agent_factory=plain_factory,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    plain_session = asyncio.run(plain_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    plain_response = asyncio.run(
        plain_adapter.prompt(
            prompt=[text_block("Inspect tools.")],
            session_id=plain_session.session_id,
        ),
    )
    assert plain_response.stop_reason == "end_turn"
    assert plain_model.last_model_request_parameters is not None
    assert [tool.name for tool in plain_model.last_model_request_parameters.function_tools] == [
        "lookup",
    ]

    prefixed_adapter = create_acp_agent(
        agent_factory=prefixed_factory,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    prefixed_session = asyncio.run(prefixed_adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    prefixed_response = asyncio.run(
        prefixed_adapter.prompt(
            prompt=[text_block("Inspect prefixed tools.")],
            session_id=prefixed_session.session_id,
        ),
    )
    assert prefixed_response.stop_reason == "end_turn"
    assert prefixed_model.last_model_request_parameters is not None
    assert [tool.name for tool in prefixed_model.last_model_request_parameters.function_tools] == [
        "repo_search",
    ]


def test_toolset_bridge_preserves_instruction_parts_and_ordering(
    tmp_path: Path,
) -> None:
    user_toolset = FunctionToolset(instructions="User capability instructions.")
    bridge_toolset = FunctionToolset(instructions=lambda: "Bridge toolset instructions.")
    model = TestModel(custom_output_text="done")

    def factory(session: AcpSessionContext) -> Agent[None, str]:
        builder = AgentBridgeBuilder(
            session=session,
            capability_bridges=[ToolsetBridge(toolset=bridge_toolset)],
        )
        contributions = builder.build(capabilities=(Toolset(toolset=user_toolset),))
        return Agent(
            model,
            capabilities=contributions.capabilities,
        )

    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    response = asyncio.run(
        adapter.prompt(
            prompt=[text_block("Inspect toolset instructions.")],
            session_id=session.session_id,
        ),
    )

    assert response.stop_reason == "end_turn"
    params = model.last_model_request_parameters
    assert params is not None
    assert params.instruction_parts is not None
    assert [(part.content, part.dynamic) for part in params.instruction_parts] == [
        ("User capability instructions.", False),
        ("Bridge toolset instructions.", True),
    ]


def test_mcp_toolset_include_instructions_reaches_model_request(tmp_path: Path) -> None:
    def return_instructions(messages: list[ModelMessage], info: AgentInfo) -> ModelResponse:
        del messages
        return ModelResponse(parts=[TextPart(info.instructions or "")])

    toolset = FunctionToolset(instructions="Be a helpful assistant.")

    @toolset.tool_plain
    def ping() -> str:
        return "pong"

    assert ping() == "pong"

    agent = Agent(
        FunctionModel(return_instructions),
        toolsets=[toolset],
    )
    adapter = create_acp_agent(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    response = asyncio.run(
        adapter.prompt(
            prompt=[text_block("Inspect MCP instructions.")],
            session_id=session.session_id,
        ),
    )

    assert response.stop_reason == "end_turn" and "Be a helpful assistant." in "".join(
        update.content.text
        for _, update in client.updates
        if getattr(update, "sessionUpdate", None) == "agent_message_chunk"
    )


def test_mcp_stdio_test_helpers_cover_script_and_env_fallbacks(tmp_path: Path) -> None:
    server_script = tmp_path / "mcp_stdio_server.py"
    _write_mcp_stdio_server_script(server_script)
    script_text = server_script.read_text(encoding="utf-8")
    assert script_text.startswith("from __future__ import annotations as _annotations")
    assert 'instructions="Be a helpful assistant."' in script_text

    existing_executable = tmp_path / "python"
    existing_executable.write_text("", encoding="utf-8")
    existing_python, existing_env = _build_mcp_stdio_test_env(
        executable=str(existing_executable),
        base_executable="/fallback-python",
        sys_path=["/repo/src", "", "/repo/tests"],
        environ={"PYTHONPATH": "/already/set"},
    )
    assert existing_python == str(existing_executable)
    assert existing_env["PYTHONPATH"] == os.pathsep.join(
        ("/repo/src", "/repo/tests", "/already/set"),
    )

    missing_python, missing_env = _build_mcp_stdio_test_env(
        executable=str(tmp_path / "missing-python"),
        base_executable="/fallback-python",
        sys_path=[""],
        environ={},
    )
    assert missing_python == "/fallback-python"
    assert "PYTHONPATH" not in missing_env

    raw_python, raw_env = _build_mcp_stdio_test_env(
        executable="/raw-python",
        base_executable=None,
        sys_path=[],
        environ={"PATH": "/usr/bin"},
    )
    assert raw_python == "/raw-python"
    assert raw_env == {"PATH": "/usr/bin"}


def test_capability_bridge_helper_and_metadata_edge_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = AcpSessionContext(
        session_id="session-2",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    agent = Agent(TestModel())
    image_bridge = ImageGenerationBridge()
    mcp_bridge = McpCapabilityBridge(
        url="https://example.com/",
        id="repo-server",
        headers={"authorization": "Bearer x"},
    )
    toolset_bridge = ToolsetBridge(toolset=cast("Any", SimpleNamespace(id=object())))
    prefix_bridge = PrefixToolsBridge(wrapped=Toolset(toolset=FunctionToolset()), prefix="repo")
    openai_bridge = OpenAICompactionBridge(message_count_threshold=4, instructions="compact")
    anthropic_bridge = AnthropicCompactionBridge(
        token_threshold=10,
        instructions="compact",
        pause_after_compaction=True,
    )

    assert _resolve_mcp_server_id("https://example.com/tools/sse", "explicit-id") == "explicit-id"
    assert _resolve_mcp_server_id("https://example.com/", None) == "example.com"
    assert _resolve_mcp_server_id("urn:acpkit", None) == "acpkit"
    assert _json_user_location(cast("Any", {"city": "Istanbul", "ignored": object()})) == {
        "city": "Istanbul",
    }

    assert len(image_bridge.build_agent_capabilities(session)) == 1
    assert len(toolset_bridge.build_agent_capabilities(session)) == 1
    assert len(prefix_bridge.build_agent_capabilities(session)) == 1
    assert len(openai_bridge.build_agent_capabilities(session)) == 1
    assert openai_bridge.get_session_metadata(session, agent) == {
        "has_trigger": False,
        "instructions": "compact",
        "message_count_threshold": 4,
    }
    assert toolset_bridge.get_session_metadata(session, agent) == {
        "toolset_id": None,
        "toolset_type": "SimpleNamespace",
    }
    assert prefix_bridge.get_session_metadata(session, agent) == {
        "prefix": "repo",
        "wrapped_capability": "Toolset",
    }
    assert prefix_bridge.get_tool_kind("search") is None
    assert mcp_bridge.get_session_metadata(session, agent)["server_id"] == "repo-server"
    assert mcp_bridge.get_tool_kind("repo.search") is None
    monkeypatch.setattr(mcp_bridge, "build_capability", lambda session: object())
    monkeypatch.setattr(anthropic_bridge, "build_capability", lambda session: object())
    assert len(mcp_bridge.build_agent_capabilities(session)) == 1
    assert len(anthropic_bridge.build_agent_capabilities(session)) == 1
    assert anthropic_bridge.get_session_metadata(session, agent) == {
        "instructions": "compact",
        "pause_after_compaction": True,
        "token_threshold": 10,
    }


def test_provider_specific_compaction_bridges_build_capabilities_and_metadata(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = AcpSessionContext(
        session_id="session-2",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    openai_bridge = OpenAICompactionBridge(
        message_count_threshold=10,
        instructions="Compact aggressively.",
    )
    anthropic_bridge = AnthropicCompactionBridge(
        token_threshold=90_000,
        instructions="Compact safely.",
        pause_after_compaction=True,
    )

    class AnthropicCompaction:
        pass

    openai_capability = openai_bridge.build_capability(session)
    assert openai_capability.get_serialization_name() == "OpenAICompaction"
    monkeypatch.setattr(
        anthropic_bridge,
        "build_capability",
        lambda session: AnthropicCompaction(),
    )
    anthropic_capability = anthropic_bridge.build_capability(session)
    assert anthropic_capability.__class__.__name__ == "AnthropicCompaction"
    assert openai_bridge.get_session_metadata(session, Agent(TestModel())) == {
        "has_trigger": False,
        "instructions": "Compact aggressively.",
        "message_count_threshold": 10,
    }


def test_capability_bridge_helper_paths_cover_import_error_fallbacks(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = AcpSessionContext(
        session_id="session-error",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    agent = Agent(TestModel())
    mcp_bridge = McpCapabilityBridge(url="https://example.com/")
    anthropic_bridge = AnthropicCompactionBridge(
        token_threshold=10,
        instructions="compact",
        pause_after_compaction=True,
    )

    monkeypatch.setattr(
        mcp_bridge,
        "build_capability",
        lambda session: (_ for _ in ()).throw(ImportError("mcp not installed")),
    )
    monkeypatch.setattr(
        anthropic_bridge,
        "build_capability",
        lambda session: (_ for _ in ()).throw(ImportError("anthropic missing")),
    )

    with pytest.raises(ImportError, match="mcp"):
        mcp_bridge.build_capability(session)
    with pytest.raises(ImportError, match="anthropic"):
        anthropic_bridge.build_capability(session)

    with pytest.raises(ImportError, match="mcp"):
        mcp_bridge.build_agent_capabilities(session)

    with pytest.raises(ImportError, match="anthropic"):
        anthropic_bridge.build_agent_capabilities(session)

    with pytest.raises(ImportError, match="anthropic"):
        anthropic_bridge.build_capability(session)

    assert anthropic_bridge.get_session_metadata(session, agent) == {
        "instructions": "compact",
        "pause_after_compaction": True,
        "token_threshold": 10,
    }


def test_anthropic_compaction_bridge_can_build_with_stubbed_import(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = AcpSessionContext(
        session_id="session-anthropic",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    bridge = AnthropicCompactionBridge(
        token_threshold=12,
        instructions="Compact.",
        pause_after_compaction=True,
    )

    class StubAnthropicCompaction:
        def __init__(
            self,
            *,
            token_threshold: int,
            instructions: str | None,
            pause_after_compaction: bool,
        ) -> None:
            self.token_threshold = token_threshold
            self.instructions = instructions
            self.pause_after_compaction = pause_after_compaction

    monkeypatch.setitem(
        sys.modules,
        "pydantic_ai.models.anthropic",
        SimpleNamespace(AnthropicCompaction=StubAnthropicCompaction),
    )

    capability = bridge.build_capability(session)
    assert isinstance(capability, StubAnthropicCompaction)
    assert capability.token_threshold == 12


def test_openai_compaction_bridge_records_visible_start_and_completion() -> None:
    session = AcpSessionContext(
        session_id="session-3",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    bridge = OpenAICompactionBridge(
        message_count_threshold=1,
        instructions="Compact aggressively.",
    )
    capability = bridge.build_capability(session)
    fake_model = cast("Any", object.__new__(_FakeOpenAIResponsesModel))
    request_context = ModelRequestContext(
        model=fake_model,
        messages=[
            ModelRequest(parts=[UserPromptPart(content="old")]),
            ModelRequest(parts=[UserPromptPart(content="new")]),
        ],
        model_settings=None,
        model_request_parameters=cast("Any", SimpleNamespace()),
    )

    updated_context = asyncio.run(
        capability.before_model_request(
            cast("Any", SimpleNamespace()),
            request_context,
        ),
    )

    assert len(updated_context.messages) == 2
    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None
    assert len(updates) == 2
    start_update = updates[0]
    progress_update = updates[1]
    assert isinstance(start_update, ToolCallStart)
    assert isinstance(progress_update, ToolCallProgress)
    assert start_update.title == "Context Compaction"
    assert start_update.status == "in_progress"
    assert start_update.raw_input == {
        "provider": "openai",
        "instructions": "Compact aggressively.",
        "message_count": 2,
    }
    assert progress_update.tool_call_id == start_update.tool_call_id
    assert progress_update.status == "completed"
    assert progress_update.raw_output == "\n".join(
        (
            "Provider: openai",
            "Status: history compacted",
            "Compaction payload stored for round-trip.",
            "Compaction id: cmp-123",
        ),
    )


class _FakeOpenAIResponsesModel(OpenAIResponsesModel):
    async def compact_messages(
        self,
        request_context: ModelRequestContext,
        *,
        instructions: str | None = None,
    ) -> ModelResponse:
        del request_context, instructions
        from pydantic_ai import CompactionPart

        return ModelResponse(
            parts=[CompactionPart(id="cmp-123", provider_name="openai", provider_details={})],
        )


class _FailingOpenAIResponsesModel(OpenAIResponsesModel):
    async def compact_messages(
        self,
        request_context: ModelRequestContext,
        *,
        instructions: str | None = None,
    ) -> ModelResponse:
        del request_context, instructions
        raise RuntimeError("boom")


def test_openai_compaction_helpers_cover_trigger_threshold_and_missing_parts() -> None:
    from pydantic_acp.bridges.capability_support import (
        _extract_compaction_part,
        _format_openai_compaction_output,
        _should_openai_compact,
    )

    messages: list[ModelMessage] = [
        ModelRequest(parts=[UserPromptPart(content="one")]),
        ModelRequest(parts=[UserPromptPart(content="two")]),
    ]

    assert (
        _should_openai_compact(
            messages,
            trigger=lambda payload: len(payload) == 2,
            message_count_threshold=None,
        )
        is True
    )
    assert (
        _should_openai_compact(
            messages,
            trigger=lambda payload: len(payload) > 10,
            message_count_threshold=0,
        )
        is False
    )
    assert (
        _should_openai_compact(
            messages,
            trigger=None,
            message_count_threshold=1,
        )
        is True
    )
    assert (
        _should_openai_compact(
            messages,
            trigger=None,
            message_count_threshold=5,
        )
        is False
    )
    assert (
        _should_openai_compact(
            messages,
            trigger=None,
            message_count_threshold=None,
        )
        is False
    )

    request_context = ModelRequestContext(
        model=cast("Any", object.__new__(_FakeOpenAIResponsesModel)),
        messages=[
            ModelRequest(parts=[UserPromptPart(content="prompt")]),
            ModelResponse(parts=[]),
            ModelResponse(
                parts=[CompactionPart(id=None, provider_name="openai", provider_details={})],
            ),
        ],
        model_settings=None,
        model_request_parameters=cast("Any", SimpleNamespace()),
    )

    compacted_part = _extract_compaction_part(request_context.messages)
    assert compacted_part is not None
    assert compacted_part.id is None
    assert _format_openai_compaction_output(request_context) == "\n".join(
        (
            "Provider: openai",
            "Status: history compacted",
            "Compaction payload stored for round-trip.",
        ),
    )

    assert (
        _extract_compaction_part([ModelRequest(parts=[UserPromptPart(content="prompt")])]) is None
    )
    assert (
        _extract_compaction_part(
            [
                ModelResponse(
                    parts=[
                        TextPart(content="skip"),
                        CompactionPart(
                            id="cmp-456",
                            provider_name="openai",
                            provider_details={},
                        ),
                    ],
                ),
            ],
        )
        is not None
    )


def test_openai_compaction_bridge_skips_when_not_needed_and_records_failures() -> None:
    session = AcpSessionContext(
        session_id="session-4",
        cwd=Path("/tmp"),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    skipped_bridge = OpenAICompactionBridge(message_count_threshold=5)
    skipped_capability = skipped_bridge.build_capability(session)
    skipped_context = ModelRequestContext(
        model=cast("Any", object.__new__(_FakeOpenAIResponsesModel)),
        messages=[
            ModelRequest(parts=[UserPromptPart(content="old")]),
            ModelRequest(parts=[UserPromptPart(content="new")]),
        ],
        model_settings=None,
        model_request_parameters=cast("Any", SimpleNamespace()),
    )

    unchanged_context = asyncio.run(
        skipped_capability.before_model_request(
            cast("Any", SimpleNamespace()),
            skipped_context,
        ),
    )

    assert unchanged_context is skipped_context
    assert skipped_bridge.drain_updates(session, Agent(TestModel())) is None

    failing_bridge = OpenAICompactionBridge(
        trigger=lambda _messages: True,
        instructions="Compact now.",
    )
    failing_capability = failing_bridge.build_capability(session)
    failing_context = ModelRequestContext(
        model=cast("Any", object.__new__(_FailingOpenAIResponsesModel)),
        messages=[
            ModelRequest(parts=[UserPromptPart(content="old")]),
            ModelRequest(parts=[UserPromptPart(content="new")]),
        ],
        model_settings=None,
        model_request_parameters=cast("Any", SimpleNamespace()),
    )

    with pytest.raises(RuntimeError, match="boom"):
        asyncio.run(
            failing_capability.before_model_request(
                cast("Any", SimpleNamespace()),
                failing_context,
            ),
        )

    updates = failing_bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None
    assert len(updates) == 2
    start_update = updates[0]
    progress_update = updates[1]
    assert isinstance(start_update, ToolCallStart)
    assert isinstance(progress_update, ToolCallProgress)
    assert start_update.title == "Context Compaction"
    assert progress_update.tool_call_id == start_update.tool_call_id
    assert progress_update.status == "failed"
    assert progress_update.raw_output == "Provider: openai\nStatus: failed\nError: boom"



from __future__ import annotations as _annotations

from .base import BufferedCapabilityBridge, CapabilityBridge
from .builtin import (
    ConfigOptionsBridge,
    DeepAgentsCompatibilityBridge,
    ModelSelectionBridge,
    ModeSelectionBridge,
    ToolSurfaceBridge,
)
from .external_hooks import EventEmissionMode, ExternalHookEventBridge

__all__ = (
    "BufferedCapabilityBridge",
    "CapabilityBridge",
    "ConfigOptionsBridge",
    "DeepAgentsCompatibilityBridge",
    "EventEmissionMode",
    "ExternalHookEventBridge",
    "ModeSelectionBridge",
    "ModelSelectionBridge",
    "ToolSurfaceBridge",
)



from __future__ import annotations as _annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Generic, TypeVar

from acp.schema import (
    AgentPlanUpdate,
    AvailableCommand,
    AvailableCommandsUpdate,
    ConfigOptionUpdate,
    CurrentModeUpdate,
    PlanEntry,
    SessionInfoUpdate,
    SessionModeState,
)
from pydantic_ai import Agent as PydanticAgent
from pydantic_ai import models as pydantic_models

from .._slash_commands import validate_mode_command_ids
from ..awaitables import resolve_value
from ..models import AdapterModel, ModelOverride
from ..providers import ModelSelectionState, ModeState
from ..session.state import AcpSessionContext, JsonValue
from ._agent_state import selected_model_id
from .session_surface import (
    ConfigOption,
    SessionSurface,
    build_mode_config_option,
    build_mode_state_from_selection,
    build_model_config_option,
    build_model_state_from_selection,
    find_model_option,
)
from .slash_commands import build_available_commands, validate_custom_commands

if TYPE_CHECKING:
    from ._session_runtime import _SessionRuntime

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

__all__ = ("_SessionSurfaceRuntime",)


class _SessionSurfaceRuntime(Generic[AgentDepsT, OutputDataT]):
    def __init__(self, runtime: _SessionRuntime[AgentDepsT, OutputDataT]) -> None:
        self._runtime = runtime

    async def build_session_surface(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> SessionSurface:
        model_selection_state = await self.get_model_selection_state(session, agent)
        mode_state = await self.get_mode_state(session, agent)
        config_options = await self.build_config_options(
            session,
            agent,
            model_selection_state=model_selection_state,
            mode_state=mode_state,
        )
        surface = SessionSurface(
            available_commands=await self.get_custom_available_commands(
                session,
                agent,
                mode_state=build_mode_state_from_selection(mode_state),
            ),
            config_options=config_options,
            model_state=build_model_state_from_selection(model_selection_state),
            mode_state=build_mode_state_from_selection(mode_state),
            plan_entries=await self.get_plan_entries(session, agent),
        )
        await self.synchronize_session_metadata(session, agent)
        self._runtime._owner._config.session_store.save(session)
        return surface

    async def emit_session_state_updates(
        self,
        session: AcpSessionContext,
        surface: SessionSurface,
        *,
        emit_available_commands: bool,
        emit_config_options: bool,
        emit_current_mode: bool,
        emit_plan: bool,
        emit_session_info: bool,
    ) -> None:
        client = self._runtime._owner._client
        if client is None:
            return
        if emit_session_info and (session.title is not None or session.metadata):
            await client.session_update(
                session_id=session.session_id,
                update=SessionInfoUpdate(
                    session_update="session_info_update",
                    title=session.title,
                    updated_at=session.updated_at.isoformat(),
                    field_meta=session.metadata or None,
                ),
            )
        if emit_available_commands:
            await client.session_update(
                session_id=session.session_id,
                update=AvailableCommandsUpdate(
                    session_update="available_commands_update",
                    available_commands=build_available_commands(
                        mode_state=surface.mode_state,
                        model_state=surface.model_state,
                        config_options=surface.config_options,
                        custom_commands=surface.available_commands,
                    ),
                ),
            )
        if emit_current_mode and surface.mode_state is not None:
            await client.session_update(
                session_id=session.session_id,
                update=CurrentModeUpdate(
                    session_update="current_mode_update",
                    current_mode_id=surface.mode_state.current_mode_id,
                ),
            )
        if emit_config_options and surface.config_options is not None:
            await client.session_update(
                session_id=session.session_id,
                update=ConfigOptionUpdate(
                    session_update="config_option_update",
                    config_options=surface.config_options,
                ),
            )
        if emit_plan and surface.plan_entries is not None:
            await client.session_update(
                session_id=session.session_id,
                update=AgentPlanUpdate(session_update="plan", entries=surface.plan_entries),
            )

    async def build_config_options(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        model_selection_state: ModelSelectionState | None,
        mode_state: ModeState | None,
    ) -> list[ConfigOption] | None:
        provider_options = await self.get_provider_config_options(session, agent)
        provider_option_ids = {option.id for option in provider_options or []}
        options: list[ConfigOption] = []
        if (
            model_selection_state is not None
            and not model_selection_state.allow_any_model_id
            and model_selection_state.enable_config_option
            and model_selection_state.current_model_id is not None
            and model_selection_state.available_models
            and "model" not in provider_option_ids
        ):
            options.append(build_model_config_option(model_selection_state))
        if (
            mode_state is not None
            and mode_state.current_mode_id is not None
            and mode_state.modes
            and "mode" not in provider_option_ids
        ):
            options.append(build_mode_config_option(mode_state))
        if provider_options is not None:
            options.extend(provider_options)
        bridge_options = self._runtime._owner._bridge_manager.get_config_options(session, agent)
        if bridge_options is not None:
            reserved_ids = {option.id for option in options}
            options.extend(option for option in bridge_options if option.id not in reserved_ids)
        return options or None

    async def get_model_selection_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> ModelSelectionState | None:
        provider = self._runtime._owner._config.models_provider
        if provider is not None:
            model_state = await resolve_value(provider.get_model_state(session, agent))
            self.synchronize_session_model_selection(session, model_state)
            return model_state
        current_model_id = self._runtime._resolve_current_model_id(session, agent)
        if current_model_id is None:
            return None
        if self._runtime._owner._config.available_models:
            return ModelSelectionState(
                available_models=list(self._runtime._owner._config.available_models),
                current_model_id=current_model_id,
                enable_config_option=self._runtime._owner._config.enable_model_config_option,
            )
        current_model_value = agent.model
        current_model_override: ModelOverride
        if isinstance(current_model_value, pydantic_models.Model | str):
            current_model_override = current_model_value
        else:
            current_model_override = current_model_id
        return ModelSelectionState(
            available_models=_default_available_models(
                current_model_id,
                current_model_value=current_model_override,
            ),
            current_model_id=current_model_id,
            enable_config_option=self._runtime._owner._config.enable_model_config_option,
        )

    async def set_provider_model_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        model_id: str,
    ) -> ModelSelectionState | None:
        provider = self._runtime._owner._config.models_provider
        if provider is None:
            return None
        model_state = await resolve_value(provider.set_model(session, agent, model_id))
        if model_state is None:
            model_state = await resolve_value(provider.get_model_state(session, agent))
        self.synchronize_session_model_selection(session, model_state)
        return model_state

    def synchronize_session_model_selection(
        self,
        session: AcpSessionContext,
        model_state: ModelSelectionState | None,
    ) -> None:
        if model_state is None:
            return
        session.session_model_id = model_state.current_model_id
        if model_state.current_model_id is None:
            session.config_values.pop("model", None)
            return
        session.config_values["model"] = model_state.current_model_id

    async def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> ModeState | None:
        provider = self._runtime._owner._config.modes_provider
        mode_state: ModeState | None = None
        if provider is not None:
            mode_state = await resolve_value(provider.get_mode_state(session, agent))
        if mode_state is None:
            mode_state = self._runtime._owner._bridge_manager.get_mode_state(session, agent)
        self.validate_mode_state(mode_state)
        self.synchronize_mode_state(session, mode_state)
        return mode_state

    async def set_provider_mode_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        mode_id: str,
    ) -> ModeState | None:
        provider = self._runtime._owner._config.modes_provider
        mode_state: ModeState | None = None
        if provider is not None:
            mode_state = await resolve_value(provider.set_mode(session, agent, mode_id))
            if mode_state is None:
                mode_state = await resolve_value(provider.get_mode_state(session, agent))
        if mode_state is None:
            mode_state = self._runtime._owner._bridge_manager.set_mode(session, agent, mode_id)
        self.validate_mode_state(mode_state)
        self.synchronize_mode_state(session, mode_state)
        return mode_state

    def validate_mode_state(self, mode_state: ModeState | None) -> None:
        if mode_state is None:
            return
        validate_mode_command_ids(mode.id for mode in mode_state.modes)

    def synchronize_mode_state(
        self,
        session: AcpSessionContext,
        mode_state: ModeState | None,
    ) -> None:
        if mode_state is None or mode_state.current_mode_id is None:
            session.config_values.pop("mode", None)
            return
        session.config_values["mode"] = mode_state.current_mode_id

    async def get_provider_config_options(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> list[ConfigOption] | None:
        provider = self._runtime._owner._config.config_options_provider
        if provider is None:
            return None
        return await resolve_value(provider.get_config_options(session, agent))

    async def get_custom_available_commands(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        *,
        mode_state: SessionModeState | None,
    ) -> list[AvailableCommand] | None:
        provider = self._runtime._owner._config.slash_command_provider
        if provider is None:
            return None
        commands = list(await resolve_value(provider.available_commands(session, agent)))
        validate_custom_commands(commands, mode_state=mode_state)
        return commands or None

    async def set_provider_config_options(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        config_id: str,
        value: str | bool,
        *,
        current_options: list[ConfigOption] | None = None,
    ) -> bool:
        provider = self._runtime._owner._config.config_options_provider
        if provider is None:
            return False
        options = await resolve_value(provider.set_config_option(session, agent, config_id, value))
        if options is not None:
            return True
        if current_options is None:
            current_options = await resolve_value(provider.get_config_options(session, agent))
        if current_options is None:
            return False
        return any(option.id == config_id for option in current_options)

    async def get_plan_entries(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> list[PlanEntry] | None:
        provider = self._runtime._owner._config.plan_provider
        if provider is None:
            return self._runtime._owner._get_native_plan_entries(session)
        plan_entries = await resolve_value(provider.get_plan(session, agent))
        return list(plan_entries) if plan_entries is not None else None

    async def synchronize_session_metadata(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        metadata_sections = self._runtime._owner._bridge_manager.get_metadata_sections(
            session,
            agent,
        )
        approval_state = await self.get_approval_state(session, agent)
        if approval_state is not None:
            metadata_sections["approval_state"] = approval_state
        plan_storage = self.plan_storage_metadata(session)
        if plan_storage is not None:
            metadata_sections["plan_storage"] = plan_storage
        preserved_metadata = {
            key: value for key, value in session.metadata.items() if key != "pydantic_acp"
        }
        if metadata_sections:
            preserved_metadata["pydantic_acp"] = metadata_sections
        session.metadata = preserved_metadata

    async def get_approval_state(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> dict[str, JsonValue] | None:
        provider = self._runtime._owner._config.approval_state_provider
        if provider is None:
            return None
        return await resolve_value(provider.get_approval_state(session, agent))

    def current_thinking_value(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> str | None:
        bridge_options = self._runtime._owner._bridge_manager.get_config_options(session, agent)
        if bridge_options is None:
            return None
        for option in bridge_options:
            if option.id == "thinking" and isinstance(option.current_value, str):
                return option.current_value
        return None

    def plan_storage_metadata(self, session: AcpSessionContext) -> dict[str, JsonValue] | None:
        if (
            self._runtime._owner._config.plan_provider is None
            and self._runtime._owner._native_plan_bridge(session) is None
        ):
            return None
        return {"directory": str(session.cwd / ".acpkit" / "plans")}

    def find_model_option(
        self,
        model_id: str,
        *,
        available_models: Sequence[AdapterModel] | None = None,
    ) -> AdapterModel | None:
        return find_model_option(
            model_id,
            available_models=available_models or self._runtime._owner._config.available_models,
        )

    def require_model_option(self, model_id: str) -> AdapterModel:
        model_option = self.find_model_option(model_id)
        if model_option is None:
            raise ValueError(model_id)
        return model_option

    def resolve_model_id_from_value(self, model_value: ModelOverride | None) -> str | None:
        model_identity = self._runtime._model_identity(model_value)
        for model_option in self._runtime._owner._config.available_models:
            if model_option.override is model_value:
                return model_option.model_id
            if model_identity is not None and model_identity == self._runtime._model_identity(
                model_option.override,
            ):
                return model_option.model_id
        if model_identity is not None:
            return model_identity
        if isinstance(model_value, str) and model_value:
            return model_value
        return None

    async def resolve_model_override(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> ModelOverride | None:
        if (
            session.session_model_id is not None
            and selected_model_id(agent) == session.session_model_id
        ):
            return None
        model_selection_state = await self.get_model_selection_state(session, agent)
        current_model_id = session.session_model_id
        model_options = self._runtime._owner._config.available_models
        if model_selection_state is not None:
            current_model_id = model_selection_state.current_model_id
            model_options = model_selection_state.available_models
        if current_model_id is None:
            return None
        model_option = self.find_model_option(
            current_model_id,
            available_models=model_options,
        )
        if model_option is not None:
            return self._runtime._resolve_model_option(model_option)
        return self._runtime._resolve_unconfigured_model_id(current_model_id)


def _known_pydantic_model_ids() -> tuple[str, ...]:
    from ._session_runtime import (
        _known_pydantic_model_ids as load_known_pydantic_model_ids,
    )

    return load_known_pydantic_model_ids()


def _known_codex_model_ids() -> tuple[str, ...]:
    from ._session_runtime import _known_codex_model_ids as load_known_codex_model_ids

    return load_known_codex_model_ids()


def _default_available_models(
    current_model_id: str,
    *,
    current_model_value: ModelOverride,
) -> list[AdapterModel]:
    available_models: list[AdapterModel] = []
    seen_model_ids: set[str] = set()

    def add_model(model_id: str, *, override: ModelOverride | None = None) -> None:
        normalized_model_id = model_id.strip()
        if not normalized_model_id or normalized_model_id in seen_model_ids:
            return
        seen_model_ids.add(normalized_model_id)
        available_models.append(
            AdapterModel(
                model_id=normalized_model_id,
                name=normalized_model_id,
                override=normalized_model_id if override is None else override,
            ),
        )

    add_model(current_model_id, override=current_model_value)
    for model_id in _known_pydantic_model_ids():
        add_model(model_id)
    for model_id in _known_codex_model_ids():
        add_model(model_id)
    return available_models
